#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import copy, functools, logging

from ttproto.core.typecheck import *
from ttproto.core.data import *
from ttproto.core.exceptions import Error
from ttproto.core.named import skip_parent_var_name
from ttproto.core import exceptions

#logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger('[packet module]')



__all__ = [
    'PacketValue',
    'PacketClass',
    'Optional',
]


class Optional:
    """Special class used to indicate that a field in a packet is optional"""

    @typecheck
    def __init__(self, type_: is_type):
        self.type = type_


class PacketValue(Value):
    """A class for representing data packets.

	A packet is a structured value containing multiple fields that are
	ordered, named and typed. The PacketValue class is typically used for
	representing protocol PDUs and their payload.

	PacketValue is an abstract class. The actual classes representing the
	PDUs are generated using the metaclass function PacketClass().

	A packet class may have multiple variants. This feature is typically
	used for protocols that define multiple kind of messages that share a
	common header.

	For instance the diagram below represents the inheritance tree for the
	ICMPv6 messages.

					 .-------.
					 | Value |
					 '-------'
					     ^
					     |
				      .-------------.
				      | PacketValue |
				      '-------------'
					     ^
					     |
			 .------------------------------------------.
			 |                  ICMPv6                  |
			 '------------------------------------------'
			      ^                 ^     ^     ^
			      |                 |     |     |
		      .------------.            |     |   .---------------------------.
		      | ICMPv6Echo |            |     |   | ICMPv6RouterAdvertisement |
		      '------------'            |     |   '---------------------------'
			^       ^               |     |
			|       |               |   .----------------------------.
	 .-------------------.  |               |   | ICMPv6NeighborSolicitation |
	 | ICMPv6EchoRequest |  |               |   '----------------------------'
	 '-------------------'  |               |
		      .-----------------.     .-----------------------------.
		      | ICMPv6EchoReply |     | ICMPv6NeighborAdvertisement |
		      '-----------------'     '-----------------------------'

	The ICMPv6 type is defined as a packet type containing 4 fields:
	 - Type		(UInt8)
	 - Code		(UInt8)
	 - Checksum	(UInt16)
	 - Body		(bytes)
	This definition is generic in the sense that it can represent any kind
	of ICMPv6 packet. However, the body onf the message is handled as a raw
	binary string, which is not convenient.

	The inheritance mechanisms of the PacketValue class allows defining
	specialised variants of the same type of message. For example, the
	ICMPv6NeighborSolicitation derives from the ICMPv6 types and has 7
	fields:
	 * Type		(UInt8)
	 * Code		(UInt8)
	 * Checksum	(UInt16)
	 - Flags	(UInt16)
	 - Reserverd	(UInt16)
	 - TargetAddress(IPv6Address)
	 - Options	(ICMPv6OptionList)
	The class inherits its first 3 fields (Type, Code & Checksum) from the
	base class ICMPv6, and defines 4 new fields in place of the binary body
	field.

	The inheritance is directly handled in the python type system. Thus we
	have the following equalities:
	  issubclass (ICMPv6NeighborSolicitation,   ICMPv6) == True
	  isinstance (ICMPv6NeighborSolicitation(), ICMPv6) == True
	  isinstance (ICMPv6(), ICMPv6NeighborSolicitation) == False

	NOTE: the latter case will always be False, even if semantically
	speaking the ICMPv6 value can be filled with an neighbour solicitation
	message (however, if well-formatted, NS messages will by default always
	be represented with the type ICMPv6Neighbor solicitation)

	Optional field
	--------------

	Some fields may not always be present in a message. Those can
	explicitely contain “no value” (which is different from containing the
	None value which indicates that the field is undefined)

	“no value” is represented with a special value class: Omit.

	When the type of a field is enclosed in the Optional class, this field
	can be indifferently field a value of this type or with the value Omit().

	For example a field containing an optional integer values will have its
	type set to “Optional(int)“.


	Default Values and tags
	-----------------------

	For each field a packet class may provide a default value.

	These default values are used when building messages. If the message is
	not fully defined (ie. some of its fields are equal to None), then the
	missing fields will be filled with their respective default value.

	Alternatively it if possible to provide a tag (inheriting from the
	PacketValue.Tag class) which can implement special processing:
	 - compute the default value of the field on a case-per-case basis
	 - making decisions when decoding a message
	For example the ttproto.core.inet.meta class provides the following tags:
	 - InetLength()		to represent the length of another field
	 - InetType()		to represent the type of another field
	 - InetVariant()	to represent the type of the current packet
	 - InetIPv6Checksum()	to compute the IPv6 checksum of a message


	The following example builds a message neighbour solicitation message
	encapsulated in an IPv6 header with all their fields left undefined
	(thus they are field with default values or automatically computed from
	a tag). The column on the right side (“default value or tag”) shows how
	the default values are defined in the class. It is not part of the
	actual output, but was added here to help understanding the process.

		>>> Message(IPv6()/ICMPv6NeighborSolicitation()).display()
		###[ IPv6 ]###					+-- default value or tag --
		  Version=                  6			| 6
		  TrafficClass=             0			| 0
		  FlowLabel=                0			| 0
		  PayloadLength=            24			| InetLength("Payload")
		  NextHeader=               58			| InetType("Payload", ...)
		  HopLimit=                 64			| 64
		  SourceAddress=            ::			| ::
		  DestinationAddress=       ::			| ::
		  Payload= 					|
		###[ ICMPv6NeighborSolicitation ]###		|
		    Type=                   135			| InetVariant(...)
		    Code=                   0			| 0
		    Checksum=               30893		| InetIPv6Checksum()
		    Reserved=               0			| 0
		    TargetAddress=          ::			| ::
		    Options= 					| []
		###[ ICMPv6OptionList ]###			+--------------------------
		Encoded as:
		    60 00 00 00 00 18 3a 40  00 00 00 00 00 00 00 00
		    00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00
		    00 00 00 00 00 00 00 00  87 00 78 ad 00 00 00 00
		    00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00

	In this example, most of the fields (Version, TrafficClass, FlowLabel,
	HopLimit, ...) just contain a default value which is copied into the
	message being built.

	Other fields have a tag indicating how to compute the default value:

	 - “PayloadLength” has the tag InetLength("Payload") which fills the
	   field with the length of the encoded payload (the neighbor
	   solicitation), whose actual length is 24 bytes.

	 - “NextHeader” has the tag InetType, which fills the field accordingly
	   to the type of the payload. In its internal dictionnary, the ICMPv6
	   type is associated to the value 58.

	 - “Type” has the tag InetVariant, which fills the field accordingly to
	   the type of the current packet (which is a variant of ICMPv6). In
	   its internal dictionnary, the ICMPv6NeighborSolicitation type is
	   associated to the value 135.

	 - “Checksum” has the tag InetIPv6Checksum(), which automatically
	   computes an IPv6 checksum from the content of the ICMPv6 messsage
	   and its enclosing IPv6 header.


	Similar processes can be implemented to customise the decoding of a
	packet, especially to make decisions for the decoder. Example:
	 - InetLength() tells what is the expected length of a field
	 - InetType() and InetVariant() tells what is the expected type of a
	   field (and thus which decoder shall be used in order to decode it
	   properly)


	TODO: Inheritance + flattening + match operations
	"""

    class Tag:
        """Base class for tags used to customise packet fields."""

        def __init__(self, default: optional(Value) = None):
            """Initialise the tag

			'default' is the default value to be used when the
			field is left undefined.
			"""
            self.__default = default

        def __repr__(self):
            return "%s(%s)" % (type(self).__name__, repr(self.__default))

        def get_default_value(self):
            """Return the default value for this field, or None if there is no default value"""
            return self.__default

        def build_message(self, value, ctx=None):
            """Build the content of the field

			This function is called by PacketValue._build_message()
			when it needs to encode the value of an individual
			field.

			'value' is the actual value of the field to be encoded.

			The default action is just to return value.build_message().


			This function may be reimplemented in order to do some
			further processing (eg. changing the presentation format, ...)

			'ctx' is meant to carry some information about the context
			in which this field is encoded. Its content is out of the
			scope of the present module. PacketValue._build_message()
			does not use it. (see ttproto.core.lib.inet.meta.InetPacketValue for
			an example of usage)
			"""
            return value.build_message()

        def decode_message(self, type_, bin_slice, ctx=None):
            """Decode the content of the field

			This function is called by PacketValue._decode_message()
			when it needs to decode the value of an individual
			field.

			'type_' is the expected type of the field.
			'bin_slice' is the slice containing the data to be decoded.

			The default action is just to return type_.decode_message(bin_slice).


			This function may be reimplemented in order to do some
			further processing (eg. changing the presentation format, ...)

			'ctx' is meant to carry some information about the context
			in which this field is encoded. Its content is out of the
			scope of the present module. PacketValue._build_message()
			does not use it. (see ttproto.core.lib.inet.meta.InetPacketValue for
			an example of usage)
			"""
            return type_.decode_message(bin_slice)

        @typecheck
        def init(self, parent: type, field_id: int, field):
            """Initialise the tag

			This function is called at the end of the constructor of
			PacketValue, once all the fields are defined.

			 - 'parent' is the PacketValue subclass that contains this field
			 - 'field_id' is the index of the field (starting from 0)
			 - 'field' is the PacketValue.Field instance representing the field
			"""
            assert parent.get_field(field_id) is field
            assert field.tag is self

            self.__packet_type = parent
            self.__field_id = field_id
            self.__field = field
            self.__default = self.__field.store_data(self.__default, none_is_allowed=True)

        def get_packet_type(self):
            """Get the PacketValue subclass that contain this field

			NOTE: this function cannot be used before init() is called.
			"""
            return self.__packet_type

        def get_field(self):
            """Get the PacketValue.Field instance representing the field
			this tag is attached to

			NOTE: this function cannot be used before init() is called.
			"""
            return self.__field

        def get_field_id(self):
            """Get the index of the field this tag is attached to

			NOTE: this function cannot be used before init() is called.
			"""
            return self.__field_id

    class Field:
        """A class representing a field in PacketValue subclasses

        A field has several attributes:
		 - name		the name of the field
		 - alias	a shortened alias for the field
		 - optional	boolean telling if the field is optional or not
		 - type		type of data that can be stored into this field
		 - tag		a tag holding a default value for the field and
		 		that can provide some special processing
		"""

        @typecheck
        def __init__(self, name: str, alias: str, typ: either(is_type, Optional), tag=None):
            """Constructor

			'typ' can be either a type (is_type()) or an instance of the Optional() class
			(to indicate that the field is optional). It is converted into a Value
			subclass using get_type()

			The 'tag' parameter can be a PacketValue.Tag class or anything else. In the
			latter case a PacketValue.Tag object is implicitely built, using this value
			as the default value of the field.
			"""

            self.name = name
            self.alias = alias
            self.optional = isinstance(typ, Optional)
            self.type = get_type(typ.type if self.optional else typ)
            if isinstance(tag, PacketValue.Tag):
                self.tag = tag
            else:
                self.tag = PacketValue.Tag(self.store_data(tag))

        def __repr__(self):
            fmt_type = "Optional(%s)" if self.optional else "%s"
            return "PacketValue.Field(%s, %s, %s, %s)" % (
                repr(self.name),
                repr(self.alias),
                fmt_type % repr(self.type),
                repr(self.tag)
            )

        @typecheck
        def store_data(self, data, none_is_allowed=True):
            """shortcut to store_data() using the Field's details (type, optional)"""

            return store_data(data, self.type, none_is_allowed=none_is_allowed, omit_is_allowed=self.optional)

    class ProxyList(list):
        """A list to ease datamanipulation prior to the construction of a PacketValue

		This class inherits from list and allows accessing its fields by their name.

		At initialisation, the list is associated with a PacketValue subclass. Then
		it is possible to use __getindex__ and __setindex__ with the
		name/alias of the field of the associated PacketValue class.

			Example:
			>>> l = PacketValue.ProxyList (IPv6)
			>>> l.append(6)		# version field
			>>> l.append(12)	# traffic class field
			>>> l.append(27)	# flow label field
			>>> l[0]
			6
			>>> l["Version"]
			6
			>>> l["tc"]
			12
			>>> l[1]
			12

		This list also allows setting up an input filter for the value
		stored into the list. By default data are stored as is. A conversion
		function can be to the constructor. This function is called each time
		an element is stored into the list, the return value of the function is
		stored in the list (in place of its input value). This can typically
		be used to implement some conversion mechanism (like calling store_data)

		The prototype of the conversion function is the following:
			def conversion_function (field: PacketValue.field, data)



		NOTE:	this class does not reimplement__delitem__(), insert(),
			pop(), remove() and sort()

		NOTE:	__getitem__ and __setitem__, when used to access a
			slice of data, do not support indexing by field name
		"""

        @typecheck
        def __init__(self, packet_type: type, convert_func: optional(callable) = None, value=None):
            def identity(field, data):
                return data

            self.set_type(packet_type)
            self.__func = convert_func if convert_func is not None else identity
            if value is not None:
                self.extend(value)

        def set_type(self, packet_type):
            assert issubclass(packet_type, PacketValue)
            self.__type = packet_type

        def get_type(self):
            return self.__type

        def clear(self):
            super()[:] = ()

        # reimplemented proxy functions
        @typecheck
        def __add__(self, other: list):
            result = copy.copy(self)
            result.extend(other)
            return result

        @typecheck
        def __iadd__(self, other: list):
            self.extend(other)
            return self

        def extend(self, other):
            for d in other:
                self.append(d)

        @typecheck
        def __imul__(self, nb: int):
            sample = copy.copy(self)
            self.clear()
            while nb > 0:
                nb -= 1
                self.extend(sample)
            return self

        def __mul__(self, nb: int):
            result = copy.copy(self)
            result *= nb
            return result

        def __rmul__(self, nb: int):
            return self * nb

        def __resolve_item(self, item: either(int, str, slice)):
            if item is None:
                return None
            elif isinstance(item, str):
                item = self.__type.get_field_id(item)
            elif isinstance(item, int):
                if item < 0:
                    item += len(self)
            else:
                assert isinstance(item, slice)
                item = slice(
                    self.__resolve_item(item.start),
                    self.__resolve_item(item.stop),
                    self.__resolve_item(item.step),
                )
            return item

        @typecheck
        def __getitem__(self, item: either(int, str, slice)):
            return super().__getitem__(self.__resolve_item(item))

        # TODO: doc warn about slices (no conversion)
        @typecheck
        def __setitem__(self, item: either(int, str, slice), value):
            item = self.__resolve_item(item)
            if isinstance(item, slice):
                return super().__setitem__(item, value)
            else:
                return super().__setitem__(item, self.__func(self.__type.get_field(item), value))

        def append(self, value):
            super().append(self.__func(self.__type.get_field(len(self)), value))

    @classmethod
    def List(cls, value=None):
        """Return a ProxyList associated to the current packet class"""
        return cls.ProxyList(cls, None, value)

    @classmethod
    def DataList(cls, value=None):
        """Return a ProxyList associated to the current packet class
		and that implement field input data conversion.

		This function is similar to .List() except that the returned list
		is initialised with a conversion that calls field.store_data()
		for each element stored into the list (thus implementing type
		conversion/initialisation for field values just like what is
		done in PacketValue's constructor.
		"""

        def store_func(field, data):
            return field.store_data(data)

        return cls.ProxyList(cls, store_func, value)

    ### class methods

    class __DescriptionProxy:
        @typecheck
        def __init__(self, func: callable):
            self.func = func

        def __getitem__(self, value):
            return self.func(value)

    @classmethod
    def metaclass_func(cls, name, bases, classdict,
                       fields: optional(list_of(tuple)) = None,
                       variant_of: optional(type) = None,
                       id: optional(int) = None,
                       prune: optional(int) = None,
                       descriptions: optional(dict_of(str, dict)) = {},
                       ):
        """Metaclass function for generating a PacketValue class.

		Parameters:

		 - fields	a list of tuples for the initialisation of the packet fields.
		 		Each tuple has the folowing content:
			 		(<name>, <alias>, <type>[, <tag_or_default_value>])

				'name' and 'alias' are string
				'type' is a core type (see is_type()) or an instance of the Optional class
				'tag'  is optional and may contain:
					- a default value
					- an instance of PacketValue.Tag (to implement special processing)

				see help (ttproto.core.packet.PacketValue.Field.__init__)

		- variant_of	Used for generating a packet class inheriting from another packet class.
		  (optional)	(thus implementing a specialised variant of the parent message). If it
		  		is not provided the new class will directly inherit from PacketValue.

				Example: ICMPv6NeighborAdvertisement can be defined as a variant_of ICMPv6

		- prune		When creating a variant of another packet, this field indicates how
		  (optional)	many fields defined in the parent class we want to keep in the new class.
		  		- If not provided, all the fields will be kept
				- If positive or zero, it indicates how many fields are kept
				- if negative (as usual in python indexing), the index will be taken
				  from the end of the list of fields. Thus, the number indicates how
				  many fields shall be removed.


				For example, the ICMPv6 message is roughly defined as follows.
					class ICMPv6 (
						metaclass = PacketClass,
						fields    = [
							("Type", 	"type",		UInt8),
							("Code",	"code",		UInt8),
							("Checksum",	"chk",		UInt16),
							("Body", 	"body",		bytes),
						]):
						pass
				This generic representation can handle any ICMPv6 message. Any
				message body can be represented since it is handled as a raw bytes
				string.

				However this format it not very convenient to manipulate (one must
				encode manually the message body. To overcome this, a specialised
				variant of packet can be derived from this class.

				Thus the neighbor solicitation message can be defined as:
					class ICMPv6NeighborSolicitation (
						metaclass  = PacketClass,
						variant_of = ICMPv6,
						prune      = -1,
						fields     = [
							("Reserved",		"rsv",	UInt32,		0),
							("TargetAddress",	"tgt",	IPv6Address,	"::"),
							("Options",		"opt",	ICMPv6OptionList, []),
						]):
					pass
				The 'prune' is set to -1 to indicate that we want to remove the last field
				of the original packet class ICMPv6 (the Body) field. Thus the resulting
				packet class ICMPv6NeighborSolicitation contains 6 fields:
				 - 3 fields inherited from the base ICMPv6 packet (Type, Code, Checksum)
				 - 3 new fields (Reserved, TargetAddress, Options)
				(also the original ICMPv6 class is not modified)

		- id		A value identifying this variant of packet
		  (optional)	Modular protocol messages typically have a “Type” which indicates which
				variant of the protocol message is present. This feature is essential
				for encoding and decoding the messages properly.

				Each hierarchy of packet classes (eg. ICMPv6 and its daughter classes)
				shares a common bi-dictionnary (of type BidictValueType) that associates
				each message variant with a value.

				This dictionnary can be accessed by adequate fields tags in order to
				encode and decode the message properly.

				See help(ttproto.core.lib.inet.meta.InetVariant) for an example.

		- descriptions	This argument may contain a dictionnary of objects providing textual
		  (optional)	descriptions in a per-field and per-value basis. This attribute	has no
				effects in the tests, its purpose is only to help understanding log
				reports.

				The object providing the description can be:
				 - a dictionnary (or any object providing __getitem__) mapping values
				   to textual descriptions
				 - a function (or any object providing __call) accepting a value as
				   parameter and returning a textual description (or None if the
				   function is unable to provide a description)

				For example, for ICMPv6 we can set a description to the 'Type' field:

				  - as a dictionnary:

					descriptions = {
					"Type": {
						1:	"Destination Unreachable",
						2:	"Packet Too Big",
					}}

				  - as a function:

				  	def icmpv6_type_description (value):
						if value == 1:
							return "Destination Unreachable"
						elif value == 2:
							return "Packet Too Big"
						else:
							return None
				  	...

					descriptions = { "Type": icmpv6_type_description }


				NOTE: when defining an inherited class (using “variant_of”) it
				if possible do overwrite the description dict for a field
				inherited from the parent class. This is especially useful when
				a common field has a different meaning in each variant of the
				message.
				For example the meaning of the “Code” field in ICMPv6 depends on
				the kind of message. Thus the ICMPv6DestinationUnreachable message
				can set:
					descriptions = {
					"Code": {
						0:	"No route to destination",
						1:	"Administratively prohibited",
						...
					}}
				without affecting the other ICMPv6 variants.
		"""

        assert len(bases) == 0  # FIXME: it could be useful to have multiple inheritance in some cases ?

        assert "variant_descriptions" not in classdict  # obsolete, use the descriptions meta-parameter instead

        if variant_of is None:
            # root class definition
            assert fields is not None
            assert prune is None

            result = type(name, (cls,), classdict)

            # initialise the dicts for resolving variant ids
            result.__cls_variants_bidict = BidictValueType(None, result)

            result.__cls_root_variant = result
            result.__cls_base_variant = None

        else:
            # variant class definition
            assert issubclass(variant_of, cls)  # FIXME: maybe too strict (subclass os PacketValue could be sufficient)

            result = type(name, (variant_of,), classdict)

            result.__cls_base_variant = variant_of

        result.__cls_variant = result
        result.__cls_variant_id = id

        if id is not None:
            assert id not in result.__cls_variants_bidict  # there should not be an existing variant with the same id

            result.__cls_variants_bidict[id] = result

        result.__init_fields(fields, prune)

        # copy the descriptions from the parent class (or initialise an empty list)
        d = variant_of.__cls_descriptions[0:result.__cls_prune] if variant_of else []
        missing = result.get_length() - len(d)
        if missing:
            d.extend([{} for v in range(0, missing)])

        # store the descriptions
        for field_name, desc_map in descriptions.items():
            i = result.get_field_id(field_name)
            # check that it provides the [] operator
            if not hasattr(desc_map, "__getitem__"):
                assert hasattr(desc_map, "__call__")  # description map must provide __getitem__ or __call__
                desc_map = cls.__DescriptionProxy(desc_map)

            # check that the dict contains only strings
            if hasattr(desc_map, "values"):
                for txt in desc_map.values():
                    assert isinstance(txt, str)
            d[i] = desc_map

        result.__cls_descriptions = d

        return result

    @classmethod
    @typecheck
    def __init_fields(cls, fields: optional(list_of(tuple)), prune: optional(int) = None):
        # check the validity of the fields
        if fields is not None:
            for f in fields:
                assert 3 <= len(f) <= 4
                assert type(f[0]) == str
                assert type(f[1]) == str
                assert is_type(f[2]) or isinstance(f[2], Optional)

        cls.__cls_payload_id = None

        if cls.__cls_base_variant is None:
            # root variant
            cls.__cls_prune = 0
            cls.__cls_fields = []
        else:
            # derived variant
            base_length = cls.__cls_base_variant.get_length()

            if prune is None:
                cls.__cls_prune = base_length
            else:
                assert -base_length <= prune <= base_length

                if prune < 0:
                    prune += base_length
                cls.__cls_prune = prune

            # keep the base payload id if not pruned
            base_pid = cls.__cls_base_variant.get_payload_id()

            if base_pid is not None and base_pid < cls.__cls_prune:
                cls.__cls_payload_id = base_pid

            # copy the fields from the base class
            cls.__cls_fields = cls.__cls_base_variant.__cls_fields[0:cls.__cls_prune]

        if fields:
            for f in fields:
                # ensure there is no name/alias collision in the field names
                for previous_field in cls.__cls_fields:
                    for name in f[0:2]:
                        assert name != previous_field.name
                        assert name != previous_field.alias

                if f[0] == "Payload":
                    assert cls.__cls_payload_id is None
                    cls.__cls_payload_id = len(cls.__cls_fields)

                cls.__cls_fields.append(PacketValue.Field(*f))

        # initialise the tags in the new fields
        i = cls.__cls_prune
        for f in cls.__cls_fields[cls.__cls_prune:]:
            f.tag.init(cls, i, f)
            i += 1

    @classmethod
    def get_length(cls):
        """Return the number of fields in this type of packet"""
        return len(cls.__cls_fields)

    @classmethod
    @typecheck
    def fields(cls) -> iterable:
        """Return an iterator on the fields of this packet (iterator
		of PacketValue.Field objects)"""

        return iter(cls.__cls_fields)

    @classmethod
    @typecheck
    def get_field(cls, id: either(int, str)) -> Field:
        """Access one individual Field either by name, alias or numerical index

		Raise exceptions.UnknownField in case the requested field does not exist
		"""
        if isinstance(id, str):
            id = cls.get_field_id(id)
        return cls.__cls_fields[id]

    @classmethod
    @typecheck
    def get_field_id(cls, arg: either(str, int)) -> int:
        """Get the numerical index of a field, given its name or numerical index

		Raise exceptions.UnknownField in case the requested field does not exist
		"""
        if isinstance(arg, int):
            assert -cls.get_length() <= arg < cls.get_length()
            return arg % cls.get_length()
        else:
            i = 0
            for f in cls.fields():
                if f.name == arg or f.alias == arg:
                    return i
                i += 1

        raise exceptions.UnknownField("Unknown field name", arg)

    @classmethod
    @typecheck
    def get_prune_id(cls) -> int:
        """Get the number of fields inherited from the parent packet class.

		If the class is not a variant of a base packet, the prune id is 0.
		"""
        return cls.__cls_prune

    @classmethod
    @typecheck
    def get_base_variant(cls) -> optional(type):
        """Return the base class for this variant of packet

		The base variant is the “variant_of” argument of PacketClass().
		It is equal to None in case the class does not inherit from
		another packet class.
		"""
        return cls.__cls_base_variant

    @classmethod
    @typecheck
    def get_root_variant(cls) -> type:
        """Return the root packet class hierarchy of PacketValue classes"""
        return cls.__cls_root_variant

    @classmethod
    def get_variants_bidict(cls) -> BidictValueType:
        """Access the bi-dictionnary that maps variant ids to variant types

		see help (PacketClass)
		"""
        # TODO: return a read-only proxy instead
        return cls.__cls_variants_bidict

    @classmethod
    @typecheck
    def get_payload_id(cls) -> optional(int):
        """Return the id of the "Payload" field.

		If there is no field named "Payload", the function just returns None.
		"""
        return cls.__cls_payload_id

    # TODO: refactor this in a cleaner way (eg. remove formats)
    @classmethod
    def get_variant_description(cls, variant_id) -> str:
        """Get a textual description of the given variant id"""

        field_id = cls.get_variant_field_id()
        txt = cls.get_description_for_value(field_id, variant_id) if field_id is not None else None

        if txt:
            return "%s %s" % (cls.get_root_variant().__name__, txt)
        elif variant_id in cls.__cls_variants_bidict:
            return cls.__cls_variants_bidict[variant_id].__name__
        elif variant_id is not None:
            return "%s Type %d" % (cls.get_root_variant().__name__, variant_id)
        else:
            return cls.get_root_variant().__name__

    @classmethod
    @typecheck
    def get_variant(cls):
        """Get the PacketValue variant of a given value.

		NOTE: the result is not necessary equal to cls (it will differ
		if cls is manually inherited from a packet class
		"""

        return cls.__cls_variant

    @classmethod
    @typecheck
    def get_variant_id(cls) -> optional(int):
        """Get the id of the variant of this class.

		The result is None if this class has no variant id set.
		"""
        return cls.__cls_variant_id

    @classmethod
    @typecheck
    def get_variant_type(cls, id: int) -> type:
        """variant id -> variant type resolution

		If the variant id is unknown, cls.get_root_variant() is
		returned. Thus this function always return a result.
		"""
        return cls.__cls_variants_bidict[id]

    @classmethod
    @typecheck
    def get_description_for_value(cls, field: either(str, int), value: is_value) -> optional(str):
        """Get the textual description for a given value in a given field

		If no description is set for this value, the function just returns None.
		"""
        try:
            return cls.__cls_descriptions[cls.get_field_id(field)][value]
        except KeyError:
            return None

    @staticmethod
    # FIXME: shall we keep this function ?
    def get_variant_field_id():
        """Return the id of the field that encodes the variant of this
		hierarchy of packet classes (or None if there is no such field)


		NOTE: the identification of the variant field id is out of the scope
		of the ttproto.core.packet module. The default implementation just returns
		None.
		"""
        return None

    ### instance methods

    @typecheck
    def __init__(self, *k, **kw):
        """Initialise a packet value.

		The function fills the content of the packet
		- either from a list of values (*k), in this case the values are expected
		  to have the same order as the
		- either by assigning the fields explicitely (**kw notation), in that case,
		  the fields are referenced by their name or alias. Unreferenced fields will
		  be left undefined.


		For example, using the ICMPv6 type definition above, the two following
		instantiations are equivalent:
			my_pkt = ICMPv6 (4, 0, 0x1234, b"   ")
			my_pkt = ICMPv6 (Type=4, Code=0, Checksum=0x1234, Body=b"   ")

		When using the explicit assignment notation, unreferenced fields are left
		undefined, thus the followings are equivalent:
			my_pkt2 = ICMPv6 (Type=4, Body=b"blah")
			my_pkt2 = ICMPv6 (4, None, None, b"blah")
			my_pkt2 = ICMPv6 (Type=4, Code=None, Checksum=None, Body=b"blah")
		"""

        Value.__init__(self)

        # TODO: copy constructor ?

        if len(k):
            if len(kw):
                raise Error('cannot mix positional and named arguments')

            missing = len(self) - len(k)
            if missing < 0:
                raise Error("to many parameters")

            self.__datas = [f.store_data(d) for d, f in zip(k, self.fields())]

            if missing:
                self.__datas.extend([None for v in range(0, missing)])
        else:
            # named parameters

            self.__datas = []
            for field in self.fields():
                for n in field.name, field.alias:
                    if n in kw:
                        v = kw.pop(n)
                        break
                else:
                    v = None

                # append the type to the internal list, but possibly convert it before
                self.__datas.append(field.store_data(v))

            if len(kw):
                raise Error('unknown fields in type %s: %s' % (type(self).__name__, ' '.join(kw.keys())))

    __len__ = get_length

    # TODO: clarify the semantics of __getitem__(type)
    @typecheck
    def __getitem__(self, index: either(int, str, is_type)) -> optional(is_data):
        """Access one field by id or name or by type (shortcut for find_type())

		The following indexing ways are supported:
		- positive or null integer
		- negative integer
		- field name
		- field alias

		Additionnaly, if a type is given as parameter, the function
		will return the result of find_type() (to look up this type
		recursively in the payload). If not found, it raises KeyError.
		"""
        if isinstance(index, str):
            index = self.get_field_id(index)
        elif isinstance(index, int):
            assert -len(self) <= index < len(self)

            if index < 0:
                index += len(self)
        else:
            assert issubclass(index, Value)
            assert self.is_flat()  # this makes things less complex (i don't think it will be necessary to support the general case)
            result = self.find_type(index)
            if result is None:
                raise KeyError
            return result

        return self.__datas[index]

    @typecheck
    def find_type(self: is_flat_value, type_: is_type) -> optional(Value):
        """Perform a recursive lookup in the payload of the packet to find a value of the requested type.

		If no value of this type is found, the function just returns None.

		Example:
			>>> Ieee802154(pl=SixLowpanIPHC(pl=IPv6(pl=UDP(pl="blah"))))
			Ieee802154(pl=SixLowpanIPHC(pl=IPv6(pl=UDP(pl='blah'))))
			>>> msg = Ieee802154(pl=SixLowpanIPHC(pl=IPv6(pl=UDP(pl="blah"))))
			>>> msg.find_type(UDP)
			UDP(pl='blah')
		"""

        pid = self.get_payload_id()
        if pid is not None:
            payload = self[pid]
            if payload is not None:
                type_ = get_type(type_)
                return payload if isinstance(payload, type_) else payload.find_type(type_)
        return None

    def __iter__(self):
        return iter(self.__datas)

    @skip_parent_var_name
    @typecheck
    def pack(self, data: is_data) -> this_class:
        """Encapsulate the given data into the present packet

		The given data is encapsulated into the "Payload" field (which
		must be present).

		If the current packet has no payload defined, this function is
		just returns self(Payload=data) (it generates a derived object
		and fills the Payload field with the given data.

		Otherwise the function will recursively call the pack() function
		on the existing payload. The result will be equal to:
			self(Payload=self["Payload"].pack(data))
		"""
        pid = self.get_payload_id()
        if pid is None:
            raise Error("Cannot pack a message into %s (no Payload field)" % type(self).__name__)

        new_seq = self()

        payloads = list(new_seq.get_datas(pid))
        field = new_seq.get_field(pid)
        if len(payloads) == 0:
            new_seq.__datas[pid] = field.store_data(data)
        elif len(payloads) >= 1:
            # FIXME: this will create lots of duplicated objects when pack is called in cascade using the __div__ operator
            #	 we may want to avoid duplicating the parent object when it is not modified
            #	 (in this case: when it is a truly anonymous object)

            # FIXME: if a sequence has multiple payloads (because of inheritance), we will
            #	 implicitely pack the data at the end of the first known payload
            #	 (this may have side effects ?)
            new_seq.__datas[pid] = field.store_data(payloads[0].pack(data))

        return new_seq

    @skip_parent_var_name
    def __truediv__(self, data):
        """Shortcut for the pack() method"""
        return self.pack(data)

    def _repr(self):
        result = []
        for field, data in zip(self.__cls_fields, self.__datas):
            if data is not None:
                result.append("%s=%s" % (field.alias, repr(data)))
        return "%s(%s)" % (type(self).__name__, ", ".join(result))

    @typecheck
    def _is_flat(self) -> bool:
        for v in self.__datas:
            if v is not None and not v.is_flat():
                return False
        return True

    @typecheck
    def get_datas(self, index: either(int, str)) -> iterable:
        """Returns an iterator on the various data defined for a given field.

		This function will examinate the current packet value and all its
		parents and yield the value of the field referenced by the
		'index' parameter if defined.

		Example:
			>>> msg = IPv6 (hl = Range (int, 0, 255))
			>>> msg2 = msg (hl = 64, nh = 58)
			>>> list(msg2.get_datas("hl"))
			[64, Range(IntValue, 0, 255)]
			>>> list(msg2.get_datas("nh"))
			[58]
			>>> list(msg2.get_datas("src"))
			[]
		"""
        seq = self
        index = self.get_field_id(index)
        while seq:
            v = seq.__datas[index]
            if v is not None:
                yield v
            seq = seq.get_parent()

    @classmethod
    @typecheck
    def _flatten(cls, seq_values: list_of(this_class)):
        # generate the new seq
        result = cls()
        for i in range(0, len(result)):
            candidates = []
            for v in seq_values:
                candidates.extend(v.get_datas(i))
            new_v = Data.flatten(*candidates)
            if new_v is not None:
                new_v.freeze()
            result.__datas[i] = new_v

        return result

    @typecheck
    def _match(self, value: this_class, mismatch_list: optional(list)) -> bool:

        assert value.is_flat()

        if value.get_variant() != self.get_variant():
            if mismatch_list is not None:
                mismatch_list.append(VariantMismatch(value, self))
            # TODO: should compare the n first field

            return False

        result = True
        i = 0
        for pattern in self.__datas:
            if pattern is not None:

                if not pattern.match(value.__datas[i], mismatch_list):
                    result = False
                    if mismatch_list is None:
                        # no need to continue
                        break
            i += 1
        return result

    def _display(self, indent, output):
        indent += 2
        import ttproto.core.list

        print("###[ %s ]###" % type(self).__name__, file=output)

        i = 0
        for f in self.fields():
            hdr = "%s%s= " % (' ' * indent, f.name)

            if f.name == "Payload":
                print(hdr,file=output)
                # TODO: display alternative payloads in a better way
                for p in self.get_datas(i):
                    p._display(indent, output)

            # FIXME: ugly hack -> do something more generic
            elif issubclass(f.type, ttproto.core.list.ListValue):
                print(hdr,file=output)
                for v in self.get_datas(i):
                    v._display(indent, output)
            else:
                datas = []
                for d in self.get_datas(i):
                    s = str(d)
                    if isinstance(d, Value):
                        desc = self.get_description_for_value(i, d)
                        if desc:
                            s = "%s (%s)" % (s, desc)
                    datas.append(s)

                print("%s%s%s" % (
                    hdr,
                    " " * (0 if len(hdr) >= 28 else 28 - len(hdr)),
                    " & ".join(datas),
                ), file=output)
            i += 1

    @typecheck
    def _fill_default_values(self: is_flat_value) -> list_of(Value):
        """Fill the undefined fields with a default value.

		This function returns a list contaning for each field:
		- the value of the field (if defined)
		- the default provided by the field tag (otherwise)
		"""
        values = []

        for f, v in zip(self.fields(), self):

            # fill the field value (and use the default if not specified)

            if v is None:
                v = f.tag.get_default_value()

                if v is None:
                    raise Error("field '%s' in '%s' must have a value" % (f.name, type(self).__name__))
                else:
                    v = f.store_data(v)
            values.append(v)

        return values

    @typecheck
    def _build_message(self) -> is_flatvalue_binary:
        """Default packet encoder

		This function fills all undefined fields with their respective
		default value, then encode these fields and concatenate them
		from left to right.
		"""

        values, bins = zip(*(f.tag.build_message(v) for v, f in zip(self._fill_default_values(), self.fields())))

        return type(self)(*values), concatenate(bins)

    @classmethod
    def _decode_message(cls, bin_slice):
        """Default packet decoder

		This function decodes the packet from left to right.
		"""

        # decode each field from left to right
        values = []


        try:
            for field in cls.fields():
                v, bin_slice = field.tag.decode_message(field.type, bin_slice)
                #log.debug('["Tag based decoder] Decoding field as: ' + str(field.type) + ' || vaue:' +str(v))
                values.append(v)
        except Exception as e:
            exceptions.push_location(e, cls, field.name)
            raise

        # create the packet
        return cls(*values), bin_slice

    def describe(self, desc):
        """Describe the packet

		desc is a MessageDescription object to be filled with the packet details.

		This function returns the description of the packet variant if
		available, otherwire it just returns  hte name of the packet
		type.

		This function always succeeds
		"""

        # TODO: maybe move this into InetPacketValue
        variant_fid = self.get_variant_field_id()
        value = None if variant_fid is None else self[variant_fid]
        desc.info = type(self).__name__ if value is None else self.get_variant_description(value)

        return True

    def describe_payload(self, desc):
        """Describe the payload of the packet if any

		desc is a MessageDescription object to be filled with the payload details.

		This function fails (returns False) if there is no payload or
		if payload.describe() fails.
		"""

        pid = self.get_payload_id()
        if pid is not None:
            pl = self[pid]
            if pl:
                return pl.describe(desc)
        return False

    def get_description(self, field_id):
        """Return the textual description of a field (if any).

		This function examines the value of the referenced field, and
		return the associated textual description if any. If there is
		no description for this value, the function just returns None.
		"""
        return self.get_description_for_value(field_id, self[field_id])


PacketClass = PacketValue.metaclass_func


@typecheck
def _diff_list_handler(pkt: PacketValue):
    i = 0
    for field, value in zip(pkt.fields(), pkt):
        yield field.alias, value, functools.partial(pkt.get_description_for_value, i)
        i += 1


DifferenceList.set_handler(PacketValue, _diff_list_handler)



#!/usr/bin/env python
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#-*- coding: iso-8859-1 -*-
################################################################################
#
# Parameter/return value type checking for Python 3000 using function annotations.
#
# (c) 2008-2010, Dmitry Dvoinikov <dmitry@targeted.org>
# Distributed under BSD license.
#
# Samples:
#
# from typecheck import *
#
# @typecheck
# def foo(i: int, x = None, s: str = "default") -> bool:
#     ...
#
# @typecheck
# def foo(*args, k1: int, k2: str = "default", k3 = None) -> nothing:
#     ...
#
# @typecheck
# def foo(ostream: with_attr("write", "flush"), f: optional(callable) = None):
#     ...
#
# divisible_by_three = lambda x: x % 3 == 0
# @typecheck
# def foo(i: by_regex("^[0-9]+$")) -> divisible_by_three:
#     ...
#
# @typecheck
# def reverse_2_tuple(t: (str, bytes)) -> (bytes, str):
#     ...
#
# @typecheck
# def reverse_3_list(t: [int, float, bool]) -> [bool, float, int]:
#     ...
#
# @typecheck
# def extract_from_dict(d: dict_of(int, str), k: tuple_of(int)) -> list_of(str):
#     ...
#
# @typecheck
# def set_level(level: one_of(1, 2, 3)):
#     ...
#
# @typecheck
# def accept_number(x: either(int, by_regex("^[0-9]+$"))):
#     ...
#
# @typecheck_with_exceptions(input_parameter_error = MemoryError):
# def custom_input_error(x: int): # now custom_input_error("foo") throws MemoryError
#     ...
#
# @typecheck_with_exceptions(return_value_error = TypeError):
# def custom_return_error() -> str: # now custom_return_error() throws TypeError
#     return 1
#
# The (6 times longer) source code with self-tests is available from:
# http://www.targeted.org/python/recipes/typecheck3000.py
#
################################################################################

__all__ = [

# decorators

"typecheck", "typecheck_with_exceptions",

# check predicates

"optional", "with_attr", "by_regex", "callable", "anything", "nothing",
"tuple_of", "list_of", "dict_of", "one_of", "either",

# exceptions

"TypeCheckError", "TypeCheckSpecificationError",
"InputParameterError", "ReturnValueError",

# utility methods

"disable",

]

################################################################################

import inspect
import functools
import re

callable = lambda x: hasattr(x, "__call__")
anything = lambda x: True
nothing = lambda x: x is None

################################################################################

_enabled = True

def disable():
    global _enabled
    _enabled = False

################################################################################

class TypeCheckError(Exception): pass
class TypeCheckSpecificationError(Exception): pass
class InputParameterError(TypeCheckError): pass
class ReturnValueError(TypeCheckError): pass

################################################################################

class Checker:

    class NoValue:
        def __str__(self):
            return "<no value>"
    no_value = NoValue()

    _registered = []

    @classmethod
    def register(cls, predicate, factory):
        cls._registered.append((predicate, factory))

    @classmethod
    def create(cls, value):
        if isinstance(value, cls):
            return value
        for predicate, factory in cls._registered:
            if predicate(value):
                return factory(value)
        else:
            return None

    def __call__(self, value):
        return self.check(value)

################################################################################

class TypeChecker(Checker):

    def __init__(self, cls):
        self._cls = cls

    def check(self, value):
        return isinstance(value, self._cls)

Checker.register(inspect.isclass, TypeChecker)

################################################################################

iterable = lambda x: hasattr(x, "__iter__")

class IterableChecker(Checker):

    def __init__(self, cont):
        self._cls = type(cont)
        self._checks = tuple(Checker.create(x) for x in iter(cont))

    def check(self, value):
        if not iterable(value):
            return False
        vals = tuple(iter(value))
        return isinstance(value, self._cls) and len(self._checks) == len(vals) and \
               functools.reduce(lambda r, c_v: r and c_v[0].check(c_v[1]),
                                zip(self._checks, vals), True)

Checker.register(iterable, IterableChecker)

################################################################################

class CallableChecker(Checker):

    def __init__(self, func):
        self._func = func

    def check(self, value):
        return bool(self._func(value))

Checker.register(callable, CallableChecker)

################################################################################

class OptionalChecker(Checker):

    def __init__(self, check):
        self._check = Checker.create(check)

    def check(self, value):
        return value is Checker.no_value or value is None or self._check.check(value)

optional = OptionalChecker

################################################################################

class WithAttrChecker(Checker):

    def __init__(self, *attrs):
        self._attrs = attrs

    def check(self, value):
        for attr in self._attrs:
            if not hasattr(value, attr):
                return False
        else:
            return True

with_attr = WithAttrChecker

################################################################################

class ByRegexChecker(Checker):

    _regex_eols = { str: "$", bytes: b"$" }
    _value_eols = { str: "\n", bytes: b"\n" }

    def __init__(self, regex):
        self._regex_t = type(regex)
        self._regex = re.compile(regex)
        self._regex_eol = regex[-1:] == self._regex_eols.get(self._regex_t)
        self._value_eol = self._value_eols[self._regex_t]

    def check(self, value):
        return type(value) is self._regex_t and \
               (not self._regex_eol or not value.endswith(self._value_eol)) and \
               self._regex.match(value) is not None

by_regex = ByRegexChecker

################################################################################

class TupleOfChecker(Checker):

    def __init__(self, check):
        self._check = Checker.create(check)

    def check(self, value):
        return isinstance(value, tuple) and \
               functools.reduce(lambda r, v: r and self._check.check(v), value, True)

tuple_of = TupleOfChecker

################################################################################

class ListOfChecker(Checker):

    def __init__(self, check):
        self._check = Checker.create(check)

    def check(self, value):
        return isinstance(value, list) and \
               functools.reduce(lambda r, v: r and self._check.check(v), value, True)

list_of = ListOfChecker

################################################################################

class DictOfChecker(Checker):

    def __init__(self, key_check, value_check):
        self._key_check = Checker.create(key_check)
        self._value_check = Checker.create(value_check)

    def check(self, value):
        return isinstance(value, dict) and \
               functools.reduce(lambda r, t: r and self._key_check.check(t[0]) and \
                                             self._value_check.check(t[1]),
                                value.items(), True)

dict_of = DictOfChecker

################################################################################

class OneOfChecker(Checker):

    def __init__(self, *values):
        self._values = values

    def check(self, value):
        return value in self._values

one_of = OneOfChecker

################################################################################

class EitherChecker(Checker):

    def __init__(self, *args):
        self._checks = tuple(Checker.create(arg) for arg in args)

    def check(self, value):
        for c in self._checks:
            if c.check(value):
                return True
        else:
            return False

either = EitherChecker

################################################################################

def typecheck(method, *, input_parameter_error = InputParameterError,
                         return_value_error = ReturnValueError):

    argspec = inspect.getfullargspec(method)
    if not argspec.annotations or not _enabled:
        return method

    default_arg_count = len(argspec.defaults or [])
    non_default_arg_count = len(argspec.args) - default_arg_count

    method_name = method.__name__
    arg_checkers = [None] * len(argspec.args)
    kwarg_checkers = {}
    return_checker = None
    kwarg_defaults = argspec.kwonlydefaults or {}

    for n, v in argspec.annotations.items():
        checker = Checker.create(v)
        if checker is None:
            raise TypeCheckSpecificationError("invalid typecheck for {0}".format(n))
        if n in argspec.kwonlyargs:
            if n in kwarg_defaults and \
               not checker.check(kwarg_defaults[n]):
                raise TypeCheckSpecificationError("the default value for {0} is incompatible "
                                                  "with its typecheck".format(n))
            kwarg_checkers[n] = checker
        elif n == "return":
            return_checker = checker
        else:
            i = argspec.args.index(n)
            if i >= non_default_arg_count and \
               not checker.check(argspec.defaults[i - non_default_arg_count]):
                raise TypeCheckSpecificationError("the default value for {0} is incompatible "
                                                  "with its typecheck".format(n))
            arg_checkers[i] = (n, checker)

    def typecheck_invocation_proxy(*args, **kwargs):

        for check, arg in zip(arg_checkers, args):
            if check is not None:
                arg_name, checker = check
                if not checker.check(arg):
                    raise input_parameter_error("{0}() has got an incompatible value "
                                                "for {1}: {2}".format(method_name, arg_name,
                                                                      str(arg) == "" and "''" or arg))

        for arg_name, checker in kwarg_checkers.items():
            kwarg = kwargs.get(arg_name, Checker.no_value)
            if not checker.check(kwarg):
                raise input_parameter_error("{0}() has got an incompatible value "
                                            "for {1}: {2}".format(method_name, arg_name,
                                                                  str(kwarg) == "" and "''" or kwarg))

        result = method(*args, **kwargs)

        if return_checker is not None and not return_checker.check(result):
            raise return_value_error("{0}() has returned an incompatible "
                                     "value: {1}".format(method_name, str(result) == "" and "''" or result))

        return result

    functools.update_wrapper(typecheck_invocation_proxy, method)

    return typecheck_invocation_proxy

################################################################################

_exception_class = lambda t: isinstance(t, type) and issubclass(t, Exception)

@typecheck
def typecheck_with_exceptions(*, input_parameter_error: optional(_exception_class) = InputParameterError,
                                 return_value_error: optional(_exception_class) = ReturnValueError):

    return lambda method: typecheck(method, input_parameter_error = input_parameter_error,
                                            return_value_error = return_value_error)

################################################################################
# EOF



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_19(CoAPTestcase):
    """Identifier:
TD_COAP_CORE_19
Objective:
Perform POST transaction with responses containing several Location-Query options
(CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 5.8.1,5.10.8,5.9.1.1

Pre-test
conditions:
•	Server accepts creation of new resource on uri  /location-query, the location of
the created resource contains two query parameters ?first=1&second=2

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a confirmable POST request to
server’s resource


2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 2 (POST)
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Content-format option
•	Empty or non-empty Payload
•	Uri-Path option "location-query"

3
Check
Server sends response containing:
•	Code = 65 (2.01 created)
•	Message ID = CMID, Token = CTOK
•	Content-format option if payload non-empty
•	Zero or more Location-path options
•	Empty or non-empty Payload

and two options of type Location-Query, with the values (none of which contains a "?" or "&"):

•	first=1
•	second=2


4
Verify
Client displays the response
"""

    def run(self):
        self.match_coap("client", CoAP(type="con", code="post",
                                       opt=self.uri(
                                           "/location-query",
                                           CoAPOptionContentFormat(),
                                       )))
        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        self.next_skip_ack()

        self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                       code=2.01,
                                       mid=CMID,
                                       tok=CTOK,
                                       opt=Opt(
                                           CoAPOptionLocationQuery("first=1"),
                                           CoAPOptionLocationQuery("second=2"),
                                       )))
        if self.match_coap("server", CoAP(pl=Not(b"")), None):
            self.match_coap("server", CoAP(
                opt=Opt(CoAPOptionContentFormat()),
            ), "fail")
        self.next_skip_ack(optional=True)



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_23(CoAPTestcase):
    """Identifier:
TD_COAP_CORE_23
Objective:
Perform PUT transaction containing the If-None-Match option (CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 5.8.1, 5.10.7,5.10.10,12.1.12

Pre-test
conditions:
•	Server should offer a /create1 resource, which does not exist and which can be created by the client
•	Client & server supports If-Non-Match


Test Sequence:
Step
Type
Description

Part A: single creation

1
Stimulus
Client is requested to send a confirmable PUT request to
server’s resource so as to atomically create the resource.

2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 3 (PUT)
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Content-format option
•	Uri-Path option "create1"
•	Option Type=If-None-Match
•	An arbitrary payload

3
Check
Server sends response containing:
•	Code = 65 (2.01 Created)
•	Message ID = CMID, Token = CTOK
•	Content-format option if payload non-empty
•	Empty or non-empty Payload


4
Verify
Client displays the response and the server created a new resource

Part B: concurrent creations

5
Stimulus
Client is requested to send a confirmable PUT request to
server’s resource so as to atomically create the resource.

6
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 3 (PUT)
•	Another client-generated Message ID ≠ CMID (➔ CMID2)
•	Client-generated Token which may or may not be ≠ CTOK (➔ CTOK2)
•	Content-format option
•	Uri-Path option "create1"
•	Option Type=If-None-Match
•	An arbitrary payload

7
Check
Server sends response containing:
•	140 (4.12 Precondition Failed)
•	Message ID = CMID2, Token = CTOK2
•	Optional Content-format option
•	Empty or non-empty Payload


8
Verify
Client displays the response
"""
    request_uri = "/create1"

    def run(self):
        # Part A

        self.match_coap("client", CoAP(type="con", code="put",
                                       opt=Opt(
                                           CoAPOptionContentFormat(),
                                           CoAPOptionUriPath(self.request_uri[1:]),
                                           CoAPOptionIfNoneMatch(),
                                       ),
                                       pl=Not(b"")))

        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        self.next_skip_ack()

        if not self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                              code=2.01, mid=CMID, tok=CTOK)):
            raise self.Stop()
        if self.match_coap("server", CoAP(pl=Not(b"")), None):
            self.match_coap("server", CoAP(
                opt=Opt(CoAPOptionContentFormat()),
            ), "fail")
        self.next_skip_ack(optional=True)

        # Part B
        self.chain()

        self.match_coap("client", CoAP(type="con", code="put",
                                       opt=Opt(
                                           CoAPOptionContentFormat(),
                                           CoAPOptionUriPath(self.request_uri[1:]),
                                           CoAPOptionIfNoneMatch(),
                                       ),
                                       pl=Not(b"")))
        CMID2 = self.frame.coap["mid"]
        CTOK2 = self.frame.coap["tok"]
        if CMID2 is Not(b''):
            if CMID2 == CMID:
                self.setverdict("fail", "Message ID should be different")
        if CTOK2 is Not(b''):
            if CTOK2 == CTOK:
                self.setverdict("fail", "Token should be different")

        self.next_skip_ack()

        if not self.match_coap("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                              code=4.12,
                                              mid=CMID2,
                                              tok=CTOK2,
                                              pl=b"")):
            raise self.Stop()

        self.next_skip_ack(optional=True)



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_13 (CoAPTestcase):
    """Identifier:
TD_COAP_CORE_13
Objective:
Perform GET transaction containing several URI-Path options (CON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 5.4.5, 5.10.2,6.5

Pre-test
conditions:
•	Server offers a /seg1/seg2/seg3 resource with resource content is not empty

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a confirmable GET request to server’s resource

2
Check
The request sent by the client contains:
•	Type = 0 (CON)
•	Code = 1 (GET)
•	Client-generated Message ID (➔ CMID)
•	Client-generated Token (➔ CTOK)
•	Uri-Path option "test"

and three options of type Uri-Path, with the values:

•	seg1
•	seg2
•	seg3


3
Check
Server sends response containing:
•	Code = 69 (2.05 content)
•	Message ID = CMID, Token = CTOK
•	Not empty Payload
•	Content format option

4
Verify
Client displays the response
"""
    def run (self):
        if self.urifilter:
            uri_path_opt = self.uri ("/seg1/seg2/seg3")
        else:
            uri_path_opt = Opt(CoAPOptionUriPath(), superset=True)

        self.match_coap ("client", CoAP (code = "get",
                        type = "con",
                        opt = uri_path_opt))

        CMID = self.frame.coap["mid"]
        CTOK = self.frame.coap["tok"]

        opts = list (filter ((lambda o: isinstance (o, CoAPOptionUriPath)), self.frame.coap["opt"]))

        if len (opts) > 1:
            self.setverdict ("pass", "multiple UriPath options")
        else:
            self.setverdict ("inconclusive", "only one UriPath option")

        # TODO: move this checks outside the testcases
        for o in opts:
            if "/" in str (o["val"]):
                self.setverdict ("fail", "option %s contains a '/'" % repr (o))

        self.next_skip_ack()

        self.match_coap ("server", CoAP (	code = 2.05,
                            type = Any (CoAPType("con"), "ack"),
                            mid=CMID,
                            tok=CTOK,
                            pl = Not (b""),
                            opt = Opt (CoAPOptionContentFormat()),
                ))





#! /usr/bin/env python
# -*- coding:utf-8 -*-

import os
from setuptools import setup, find_packages

PACKAGE = 'ttproto'
LICENSE = 'CeCILL'

# Read version without importing for coverage issues
def get_version(package):
    """ Extract package version without importing file

    Importing module alter coverage results and may import some non-installed
    dependencies. So read the file directly

    Inspired from pep8 setup.py
    """
    with open(os.path.join(package, '__init__.py')) as init_fd:
        for line in init_fd:
            if line.startswith('__version__'):
                return eval(line.split('=')[-1])  # pylint:disable=eval-used


CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Topic :: Internet",
    "Topic :: Software Development :: Testing",
    "Topic :: Scientific/Engineering",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS"
]

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name=PACKAGE,
    version=get_version(PACKAGE),
    url='https://www.irisa.fr/tipi/wiki/doku.php/testing_tool_prototype',
    author='Universite de Rennes 1 / INRIA',
    maintainer='Federico Sismondi',
    maintainer_email='federico.sismondi@inria.fr',
    description=('ttproto is an experimental tool for implementing testing'
                 'tools, for conformance and interoperability testing mainly.'),
    license=LICENSE,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=CLASSIFIERS,
    install_requires=[
        'pyyaml',
        'requests',
        'pika',
    ],

)



#!/usr/bin/env python3

from ttproto.ts_coap.proto_specific import CoAPTestcase
from ttproto.ts_coap.proto_templates import *


class TD_COAP_CORE_17 (CoAPTestcase):
    """Identifier:
TD_COAP_CORE_17
Objective:
Perform GET transaction with a separate response (NON mode)
Configuration:
CoAP_CFG_BASIC
References:
[COAP] 2.2, 5.2.2,  5.8.1

Pre-test
conditions:
•	Server offers a resource /separate which is not served immediately and which therefore is not acknowledged in a piggybacked way.

Test Sequence:
Step
Type
Description

1
Stimulus
Client is requested to send a non-confirmable GET request to server’s resource

2
Check
The request sent by the client contains:
•	Type = 1 (NON)
•	Code = 1 (GET)
•	Client-generated Message ID (➔ CMID)
•	Uri-Path option "separate"

3
Check
Server DOES NOT send response containing:
•	Type = 2 (ACK)
•	Same message ID as in the request in step 2
•	empty Payload


4
Check
Server sends response containing:
•	Type  = 1 (NON)
•	Code = 69 (2.05 content)
•	Server-generated Message ID (➔ SMID)
•	Content-format option
•	Non-empty Payload

5
Verify
Client displays the response
"""
    def run (self):
        self.match_coap ("client", CoAP (type="non", code = "get",
                    opt=self.uri("/separate")))

        self.next()

        #FIXME: may be out-of-order
        if self.frame.coap in CoAP (type="ack"):
            self.setverdict ("fail", "server must no send any ack")
            self.next()

        if self.match_coap ("server", CoAP (type="non", code=2.05)):
            self.match_coap ("server", CoAP (
                        pl = Not (b''),
                        opt= Opt(CoAPOptionContentFormat())
                ), "fail")





#!/usr/bin/env python3
#
#  (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re

from .templates import *
from ttproto.core.analyzer import TestCase, is_protocol, Node, Conversation, Capture
from ttproto.core.dissector import Frame
from ttproto.core.templates import All, Not, Any, Length
from ttproto.core.typecheck import *
from ttproto.core.lib.all import *
from urllib import parse
from ttproto.core.exceptions import Error

from xml.etree import ElementTree
import logging
import requests
import json
import sys

# CoAP constants
RESPONSE_TIMEOUT = 2
RESPONSE_RANDOM_FACTOR = 1.5
MAX_RETRANSMIT = 4
MAX_TIMEOUT = 10 + round(
    (RESPONSE_TIMEOUT * RESPONSE_RANDOM_FACTOR) * 2 ** MAX_RETRANSMIT
)


class NoStimuliFoundForTestcase(Error):
    """
    Error raised when no stimuli was defined for the testcase
    """
    pass


class FilterError(Error):
    """
    The error thrown when an error occurs into Filter object during its run
    """
    pass


class CoAPTestCase(TestCase):
    """
    The test case extension representing a CoAP test case
    """

    # Some default parameters
    reverse_proxy = False
    urifilter = False

    @classmethod
    @typecheck
    def get_protocol(cls) -> is_protocol:
        """
        Get the protocol corresponding to this test case. This has to be
        implemented into the protocol's common test case class.

        :return: The protocol on which this TC will occur
        :rtype: Value
        """
        return CoAP

    @classmethod
    @typecheck
    def get_nodes_identification_templates(cls) -> list_of(Node):
        """
        Get the nodes of this test case. This has to be be implemented into
        each test cases class.

        :return: The nodes of this TC
        :rtype: [Node]

        .. note:: For CoAP it is simpler so we can define this function in this
                  class but for other protocols it can happend that we have to
                  define this inside each TC
        """
        return [
            Node('client', UDP(dport=5683)),
            Node('server', UDP(sport=5683))
        ]

    @typecheck
    def preprocess(
            self,
            capture: Capture
    ) -> (list_of(Conversation), list_of(Frame)):
        """
        Preprocess and filter the frames of the capture into test case related
        conversations.

        :param Capture: The capture which will be filtered/preprocessed
        :return: list of conversations and list of ignored frames
        """

        # TODO assert is subclass of TesCase?

        # Get informations from the test case
        # TODO: Get attrbutes stimuli , protocol under test, nodes patterns
        #       directly from child's atrib?

        stimulis = self.get_stimulis()
        protocol = self.get_protocol()
        nodes = self.get_nodes_identification_templates()

        conversations = []
        ignored = []

        # TODO what happens if no protocol declared on the test case?
        if not nodes or len(nodes) < 2:
            raise ValueError(
                'Expected at leaset two nodes declaration from the test case'
            )
        if not protocol:
            raise ValueError(
                'Expected a protocol under test declaration from the test case'
            )

        # If there is no stimuli at all
        if not stimulis or len(stimulis) == 0:
            raise NoStimuliFoundForTestcase(
                'Expected stimuli declaration from the test case'
            )

        # Get the frames related with the protocol under test
        frames, ignored = Frame.filter_frames(capture.frames, protocol)

        # Get a counter of the current stimuli
        sti_count = 0
        current_conversation = None
        nb_stimulis = len(stimulis)
        for frame in frames:

            # If the frame matches a stimuli
            if stimulis[sti_count].match(frame[protocol]):

                # If it's the first stimuli
                if sti_count == 0:

                    # If there is already a conversation pending, save it
                    if current_conversation:
                        self._conversations.append(current_conversation)

                    # Get the nodes as a list of nodes
                    # TODO already done at begining. why isnde the iteeration?
                    # nodes = testcase.get_nodes_identification_templates()

                    # And create the new one
                    current_conversation = Conversation(nodes)

                # If intermediate stimulis, just increment the counter
                sti_count = (sti_count + 1) % nb_stimulis

            # If there is a current_conversation, put the frame into it
            if current_conversation:
                current_conversation.append(frame)

            # If no conversation pending
            else:
                ignored.append(frame)

        # At the end, if there is a current conversation pending, close it
        if current_conversation:

            # If not all stimulis were consumed
            if sti_count != 0:
                raise FilterError(
                    'Not all stimulis were consumed, %d left and next one should have been %s'
                    %
                    (
                        nb_stimulis - sti_count,
                        stimulis[sti_count]
                    )
                )

            # Close the current conversation by adding it to list
            conversations.append(current_conversation)

        return conversations, ignored

    @typecheck
    def next_skip_ack(self, optional: bool = False):
        """
        Call self.next() but skips possibly interleaved ACKs

        :param optional: If we have to get a next frame or not
        :type optional: bool
        """

        # Goes to next frame
        self.next(optional)

        # While there is one and that it's an ack, pass it
        while all((
                    self._frame is not None,
                    self._frame[CoAP] in CoAP(type='ack', code=0)
        )):
            self.next(optional)

    @property
    def coap(self) -> CoAP:
        """
        Get the coap layer of the current frame

        :return: The coap layer of the current frame
        :rtype: CoAP
        """
        return self._frame[CoAP]

    @typecheck
    def uri(self, uri: str, *other_opts):
        """
        Filter for disabling a template if URI-Filter is disabled

        :param uri: The uri
        :param other_opts: More options
            Elemements may be either:
                - CoAPOption datas
                    -> will be fed into a Opt() together with the Uri options
                - CoAPOptionList datas
                    -> will be combined with the Opt() within a All() template
        :type uri: str
        :type other_opts: tuple of more parameters
        """
        opt = []
        opt_list = []
        for o in other_opts:
            if issubclass(o.get_type(), CoAPOption):
                opt.append(o)
            elif issubclass(o.get_type(), CoAPOptionList):
                opt_list.append(o)
            else:
                raise ValueError

        if self.urifilter:
            u = urllib.parse.urlparse(uri)
            if u.path:
                assert not any(
                    isinstance(v, CoAPOptionUriPath) for v in other_opts
                )
                for elem in u.path.split("/"):
                    if elem:
                        opt.append(CoAPOptionUriPath(elem))
            if u.query:
                assert not any(
                    isinstance(v, CoAPOptionUriQuery) for v in other_opts
                )
                for elem in u.query.split("&"):
                    if elem:
                        opt.append(CoAPOptionUriQuery(elem))

        if opt:
            opt_list.append(Opt(*opt))

        if not opt_list:
            return None
        elif len(opt_list) == 1:
            return opt_list[0]
        else:
            return All(*opt_list)


class Link(list):
    """
    Class representing the link values for CoAP

    .. example:: coap://[adress][/uri]?[par_name]=[par_token]
    .. note:: In the whole class, the string representation of the link value
              got by parsing the payload of the CoAP packet is stored into the
              's' variable
    """

    __re_uri = re.compile(r"<([^>]*)>")
    __re_par_name = re.compile(r";([0-9A-Za-z!#$%&+^_`{}~-]+)(=?)")
    __re_ptoken = re.compile(r"[]!#$%&'()*+./:<=>?@[^_`{|}~0-9A-Za-z-]+")

    class FormatError(Exception):
        """
        Error thrown when there is a format error during parsing the uri value
        """
        pass

    @typecheck
    def is_compiled_regex(arg):
        """
        Check if a parameter is a valid regex compiled object.
        This function is used for the typechecker decorator.

        :return: True if a valid compiled regex, False if not
        :rtype: bool
        """
        return all((
            arg is not None,
            isinstance(arg, type(re.compile('dummy_pattern')))
        ))

    @typecheck
    def __init__(self, pl: bytes):
        """
        Initialize the link representation

        :param pl: The payload of the CoAP layer
        :type pl: bytes
        """

        @typecheck
        def error(msg: str, in_string=None):
            """
            Function to manage format errors

            :param msg: The string that we were seeking
            :param in_string: The object in which we were seeking
            :type msg: str
            :type in_string: anything

            :raises FormatError: Always raises this exception because it's the
                                 main purpose of this function
            """
            if in_string:
                raise self.FormatError("%s in %r" % (msg, in_string))
            else:
                raise self.FormatError("%s at %r..." % (msg, s[:40]))

        @typecheck
        def have(string: str) -> bool:
            """
            Check if the link value contains the entered parameter

            :param string: The string to seek in the link value
            :type string: str

            :return: True if the string is contained into the link value
            :rtype: bool
            """
            return s and s.startswith(string)

        @typecheck
        def percent_unquote(string: str) -> str:
            """
            Replace '%xx' url encoded special characters by their string value

            :param string: The uri value from which we will replace those
            :type string: str

            :raises FormatError: If the urllib was unable to replace the '%xx'
                                 characters

            :return: The uri with the special characters as normal char
            :rtype: str
            """
            try:
                return urllib.parse.unquote(string, errors='strict')
            except UnicodeDecodeError as e:
                error(str(e), string)

        # Try to parse the given payload into utf-8 string representation
        try:
            s = str(pl, 'utf-8')
        except UnicodeDecodeError as e:
            error(str(e), pl)

        # Match object used here
        mo = None

        @typecheck
        def consume(pattern: either(str, is_compiled_regex), subject: str):
            """
            Consume a part of the link value (stored into s variable)

            :param pattern: The pattern to consume
            :param subject: The name of the subject to consume
            :type pattern: either(str, _sre.SRE_Pattern)
            :type subject: str
            """

            # Get the extern s (for uri current value) and mo (match object)
            nonlocal s, mo

            # If the pattern is a string
            if isinstance(pattern, str):
                if s.startswith(pattern):
                    s = s[len(pattern):]
                    return

            # If the pattern is a compiled regex
            else:
                mo = re.match(pattern, s)
                if mo:
                    s = s[mo.end(0):]
                    return

            # If it didn't match with the current uri value
            error("malformed %s" % subject)

        # If it managed to get a string (uri value) from the payload
        if s:
            while True:

                # Link-value
                consume(self.__re_uri, 'uri')

                # Get the uri value without the encoded special chars
                uri = percent_unquote(mo.group(1))

                # Store the value of it
                link_value = self.LinkValue(uri)

                # While there is a ";" we can process the parameter
                while have(";"):

                    # Link-param
                    consume(self.__re_par_name, 'parmname')

                    # Get the name of the parameter
                    name = mo.group(1)

                    # If no value associated to it
                    if not mo.group(2):
                        value = None

                    # If it does have a value associated
                    elif have('"'):

                        # Quoted-string => Read and unquote it
                        value = []
                        esc = False

                        # Parse the whole uri left
                        for i in range(1, len(s)):

                            # Get each character
                            c = s[i]

                            # If it's not escaped
                            if not esc:

                                # If it's escaped now
                                if c == '\\':
                                    esc = True

                                # End of string
                                elif c == '"':
                                    break

                                # Other chars
                                # TODO: Normalise LWS
                                else:
                                    value.append(c)

                            # If it is escaped
                            else:
                                esc = False

                                # Quoted char
                                if c == '"' or c == '\\':
                                    value.append(c)

                                # Was an unquoted \
                                else:
                                    value.append('\\' + c)

                        # Error: unterminated quoted-string
                        else:
                            error(
                                "attribute value for %r is %s"
                                %
                                (
                                    name,
                                    'an unterminated quoted-string'
                                )
                            )

                        # Transform the value char list into a single string
                        value = ''.join(value)

                        # If still empty
                        if not value:
                            error("attribute value for %r is empty" % name)

                        # Consume the read part
                        s = s[i + 1:]

                    # If it doesn't begin with a quote, it's a token
                    else:

                        # Consume it
                        consume(self.__re_ptoken, 'ptoken')

                        # And generate its value
                        value = percent_unquote(mo.group(0))

                    # Add the link value pair
                    link_value.append((name, value))

                # In the end, append the link value to the link list
                self.append(link_value)

                # If we finished reading the whole uri, s is now empty
                if not s:
                    break

                # Next link-value
                consume(",", "delimiter, expected ','")

    class LinkValue(list):
        """
        A class representing the link values which consist into a parameter
        name associated a parameter value
        """

        @typecheck
        def __init__(self, uri: str):
            """
            Initialize the link value object

            :param uri: The basic uri
            :type uri: str
            """
            self.uri = uri

        @typecheck
        def get(
                self,
                par_name: str,
                testcase: optional(CoAPTestCase) = None
        ) -> optional(str):
            """
            Get the value of a link value from its parameter name

            :param par_name: The parameter name
            :param testcase: The TestCase object to put its verdict to 'fail'
            :type par_name: str
            :type testcase: optional(CoAPTestCase)

            :return: The parameter value associated to this parameter name if
                     one found, None if none found
            :rtype: optional(str)
            """

            # The result to return, None at the beginning to check if a link
            # value has multiple values for a single parameter name
            result = None

            # For each couple (name => value) inside this list
            for name, value in self:

                # If the parameter name is found
                if name == par_name:

                    # If no result found until here, ok and put it
                    if result is None:
                        result = value

                    # If a value was already put and another one if found
                    else:
                        msg = (
                            "link-value contains multiple %r parameters"
                            %
                            par_name
                        )

                        # If a test case object is given, its verdict fails
                        if testcase:
                            testcase.set_verdict('fail', msg)
                        else:
                            raise Exception(msg)

            # Return the value if one found, if none found just return None
            return result


def validate(jsond, objectid):
    logging.info("ObjectID: " + objectid)
    jsondata = jsond[2:]
    jsondata = jsondata[:-1]
    jsondata = json.loads(jsondata)
    # logging.info json.dumps(jsondata, sort_keys=True,indent=4, separators=(',', ': '))

    url = 'http://www.openmobilealliance.org/api/lwm2m/v1/Object?ObjectID=' + objectid;
    logging.info("Registry url: " + url)
    r = requests.get(url)
    # logging.info json.dumps(r.json(), sort_keys=True,indent=4, separators=(',', ': '))
    j = r.json()
    logging.info("ObjectLink: " + j[0]["ObjectLink"])
    r2 = requests.get(j[0]["ObjectLink"])
    # logging.info r2.content
    root = ElementTree.fromstring(r2.content)

    validation = "pass"
    for item in root.iter('Item'):
        for child in item:
            if (child.text == "Mandatory"):
                if (item.find("Operations").text != "E"):
                    logging.info("ResourceID: " + item.get("ID"))
                    logging.info("Name: " + str(item.find("Name").text))
                    logging.info("Type: " + str(item.find("Type").text))
                    logging.info("Operations: " + str(item.find("Operations").text))
                    logging.info("MultipleInstances: " + str(item.find("MultipleInstances").text))
                    for attrs in jsondata['e']:
                        itemuri = item.get("ID");
                        if (item.find("MultipleInstances").text == "Multiple"):
                            itemuri = itemuri + "/0";
                        if attrs['n'] == itemuri:
                            n = attrs['n']
                            logging.info("Found!")
                            break
                    else:
                        validation = "fail"
                        logging.info('Not found!')
    logging.info("Validation: " + validation)
    return validation


if __name__ == "__main__":
    pass



from ..common import *


class TD_6LOWPAN_ND_HC_01 (SixlowpanTestCase):
    """
---
TD_6LOWPAN_ND_HC_01:
    cfg: 6ln to 6lr
    not: Null
    obj: check that EUTs are able to perform the neighbor Discovery (EUI-64
         link-local, hop limit=255)
    pre:
         - Header compression is enabled on both EUT1 and EUT2.
         - EUT1 and EUT2 are configured to use EUI-64 addresses.
         - EUT1 is configured as 6LN.
         - EUT2 is configured as 6LR.
    ref: RFC 6282 section 3; RFC 6775 5.6
    seq:
        -   s:
               - Initialize the network interface of the 6LR (EUT2)
               - Initialize the network interface of the 6LN (EUT1)
               - Hop Limit is 255, no traffic class or flow label is being used
        -   c: The 6LN sends à Neighbor Solicitation (NS) to all-nodes multicast address with SLLAO (EUI-64).
               Source = link-local based on EUI-64
        -   c: The 6LR (EUT2) receives the Neighbor Solicitation (NS) the 6LN (EUT1)
        -   f: "Hop Limit is 64 and source address begin by fe80::"
        -   f: In IP_HC, TF is 11 and the ecn, dscp and flow label fields are
                compressed away
        -   f: In IP_HC, HLIM (HL) is 10 and the hop limit field is compressed
                away
        -   f: In IP_HC, SAC=0, SAM=11; DAC=0; DAM=01
        -   v: The 6LR (EUT2) receives the Neighbor Solicitation (NS) the 6LN (EUT1)
        -   c: The 6LR (EUT2) sends a unicast Neighbor Advertisement (NA) containing the ARO.
               Link local addresses are used.
        -   f: "Hop Limit is 1 and source address begin by fe80::"
        -   f: In IP_HC, TF is 11 and the ecn, dscp and flow label fields are
                compressed away
        -   f: In IP_HC, HLIM (HL) is 10 and the hop limit field is compressed
                away
        -   f: In IP_HC, SAC=0, SAM=11; DAC=0; DAM=11
        -   v: The 6LN (EUT1) receives the Neighbor Advertisement (NA) from the 6LR (EUT2)
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. warning::
            For the moment, we didn't manage to generate packets with the
            wanted size so we just don't take them into account
        """
        return [
            SixLowpanIPHC(
                pl=All(
                    # Length(IPv6, 35),
                    IPv6(
                        tc=0x00,
                        fl=0x00,
                        hl=255,
                        pl=ICMPv6NeighborSolicitation(
                            # pl=Length(bytes, 0)
                        )
                    )
                )
            )
        ]

    @classmethod
    @typecheck
    def get_nodes_identification_templates(cls) -> list_of(Node):
        """
        Get the nodes of this test case. This has to be be implemented into
        each test cases class.

        :return: The nodes of this TC
        :rtype: [Node]

        """
        return [
            Node('EUT1', ICMPv6NeighborSolicitation()),
            Node('EUT2', ICMPv6NeighborAdvertisement())
        ]

    def run(self):
        # NOTE: Should we check the IP adresses to check that it is really the
        #       EUT1 and EUT2?

        # TS 1
        self.match('EUT1', SixLowpanIPHC(pl=IPv6(pl=ICMPv6NeighborSolicitation())))

        # TS 2
        self.match('EUT1', SixLowpanIPHC(dp=0b011))

        # TS 3
        self.match('EUT1', SixLowpanIPHC(
            tf=0b11,
            iecn=0b00,
            sci=Omit(),
            idscp=0b00,
            ifl=0x00
        ))
        self.match('EUT1', SixLowpanIPHC(pl=IPv6 (HopLimit = 255)))

        # TS 4
        self.match('EUT1', SixLowpanIPHC(hl=0b11, ihl=Omit()))

        # TS 5
        self.match('EUT1', SixLowpanIPHC(
            sac=False,
            sam=0b11,
            dac=False,
            dam=0b01
        ))

        # TS 6
        # NOTE: Only one sniff file so we can't check that the EUT2 didn't
        #       receive the echo request message

        self.next()

        # TS 7
        self.match('EUT2', SixLowpanIPHC(pl=IPv6(pl=ICMPv6NeighborAdvertisement())))

        # TS 8
        self.match('EUT2', SixLowpanIPHC(dp=0b011))

        # TS 9
        self.match('EUT2', SixLowpanIPHC(
            tf=0b11,
            iecn=0b00,
            sci=Omit(),
            idscp=0b00,
            ifl=0x00
        ))

        # TS 10
        self.match('EUT2', SixLowpanIPHC(hl=0b11, ihl=Omit()))

        # TS 11
        self.match('EUT2', SixLowpanIPHC(
            sac=False,
            sam=0b11,
            dac=False,
            dam=0b11
        ))

        # TS 12
        # NOTE: Only one sniff file so we can't check that the EUT2 didn't
        #       receive the echo request message



from ..common import *


class TD_COAP_CORE_17 (CoAPTestCase):
    """
---
TD_COAP_CORE_17:
    cfg: CoAP_CFG_BASIC
    obj: Perform GET transaction with a separate response (NON mode)
    pre: Server offers a resource /separate which is not served immediately
        and which therefore is not acknowledged in a piggybacked way.
    ref: '[COAP] 2.2, 5.2.2, 5.8.1'
    seq:
    -   s: "Client is requested to send a non-confirmable GET request to\
            \ server\u2019s resource"
    -   c:
        - 'The request sent by the client contains:'
        -   - Type = 1 (NON)
            - Code = 1 (GET)
            - "Client-generated Message ID (\u2794 CMID)"
            - Uri-Path option "separate"
    -   c:
        - 'Server DOES NOT send response containing:'
        -   - Type = 2 (ACK)
            - Same message ID as in the request in step 2
            - Empty Payload
    -   n: Some time (a couple of seconds) elapses.
    -   c:
        - 'Server sends response containing:'
        -   - Type = 1 (NON)
            - Code = 2.05 (Content)
            - "Server-generated Message ID (\u2794 SMID)"
            - Content-format option
            - Non-empty Payload
    -   v: Client displays the response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. note::
            Check the number/value of the uri query options or not?
        """
        return [
            CoAP(type='non', code='get')
        ]

    def run (self):
        self.match ("client", CoAP (type="non", code = "get",
                    opt=self.uri("/separate")))

        self.next()

        #FIXME: may be out-of-order
        if self.coap in CoAP (type="ack"):
            self.set_verdict ("fail", "server must no send any ack")
            self.next()

        if self.match ("server", CoAP (type="non", code=2.05)):
            self.match ("server", CoAP (
                        pl = Not (b''),
                        opt= Opt(CoAPOptionContentFormat())
                ), "fail")





import unittest

from ttproto.core.analyzer import Node
from ttproto.core.data import Value
from ttproto.core.typecheck3000 import InputParameterError
from ttproto.core.lib.all import *


class NodeTestCase(unittest.TestCase):
    """
    Test class for the node class
    """

    # #################### Init and deinit functions #########################
    def setUp(self):
        """
            Initialize the analyzer instance
        """
        self.node = Node('name', UDP(sport=5683))

    # #################### Tests functions #########################

    # ##### __init__
    def test___init__(self):

        # Normal use
        self.node = Node('name', UDP(sport=5683))

        # Wrong name
        with self.assertRaises(InputParameterError):
            self.node = Node(IPv4(src='127.0.0.1'), UDP(sport=5683))

        # Wrong value
        with self.assertRaises(InputParameterError):
            self.node = Node('name', '127.0.0.1')

    # ##### name
    def test_name(self):

        # Get name
        name = self.node.name
        self.assertIsInstance(name, str)

        # Set name
        with self.assertRaises(AttributeError):
            self.node.name = name

    # ##### value
    def test_value(self):

        # Get value
        value = self.node.value
        self.assertIsInstance(value, Value)

        # Set value
        with self.assertRaises(AttributeError):
            self.node.value = value


# #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# -*- coding: utf-8 -*-

"""This module implement the snapshot semantics of ttproto

"""

import numbers, threading, types

from	ttproto.core.typecheck	import *
from	ttproto.core		import clock, exceptions, logger, named

from	contextlib	import contextmanager

# TODO: split into several modules
_all__ = [
	'SnapshotBlock',
	'AltBlock',
	'SnapshotManager',
	'repeat',
	'SnapshotContext',
	'alt',
	'EventSource',
	'Event',
	'EventMatch',
	'Timer',
	'EventTimerStarted',
	'EventTimerStopped',
	'EventTimerTimeout',
	'Altstep',
	'AltstepBlock',
	'altstep',
]


# TODO: document subtilities
#
#	@altstep
#	def match_timeout (*timer_list):
#
#		for timer in timer_list:
#			@timer.timeout
#			def _():
#				print "timer timeout !", timer
#				repeat()
#   -> invalid use of free variables in the handler function
#      here, 'timer' will always refer to the last timer in the list
#      because both events branchs are instantiated in the same frame
#

class SnapshotBlock:
	"""A snapshot block holds a list of branches that can be followed


	"""
	#FIXME: Block may not be the most adequate name

	@typecheck
	def __init__ (self, name: optional (str) = None):
		self.branch_list = []
		self.__parent = None
		self.__name = name

	@typecheck
	def stack (self) -> iterable:
		'''returns an iterator on the block stack'''

		while self:
			yield self
			self = self.__parent

	@typecheck
	def push (self, block: this_class) -> this_class:
		'''returnes the new base of the stack'''

		assert block.__parent is None
		block.__parent = self
		return block

	@typecheck
	def pop (self) -> this_class:
		'''returnes the new base of the stack'''
		result = self.__parent
		self.__parent = None
		return result

	@typecheck
	def get_sources (self) -> iterable:
		for g in self.branch_list:
			for s in g.event.get_sources():
				yield s
		if self.__parent:
			for s in self.__parent.get_sources():
				yield s

	def match (self)  -> "optional (EventMatch)":
		raise NotImplementedError()


class AltBlock (SnapshotBlock):

	def match (self)  -> "optional (EventMatch)":

#		print "enter alt match"
		for block in self.stack():
#			print "  block", block
			if isinstance (block, AltBlock):
				# evaluate the conditions
				for g in block.branch_list:
#					print "  try", g
					em = g.match()
					if em:
						# matched the event!
#						print "leave alt match"
						em.set_branch (g)
						return em
#		print "leave alt match"
		return None

#TODO: the 'editable' concept is not very clear -> document clearly when a block is no longer editable and what are the implications
class SnapshotManager:
	"""SnapshotManager is the main class controling the snapshot semantics.

	There is one SnapshotManager object associated to each thread
	performing snapshot operations. These instances are created
	automatically the first time the manager object is needed by the
	thread.

	"""

	__thread_local = threading.local()


	# Key: 	 SnapshotManager object
	# Value: True if the thread is running
	#	 False if the thread is awaiting external events
	__tracked_managers = {}
	__tracked_lock = threading.Lock()


	class Repeat (BaseException):
		"""Repeat Exception

		This exception notifies the snapshot manager that the current
		snapshot block should be reevaluated again.

		It can can be raised by:
		- an Event (in its test function: match_xxxxxxx())
		- a branch handler executed upon an event is match
		- a snapshot block being matched
		"""
		pass

	class Sources:
		"""A class for listing the event sources that are grabbed by a snapshot manager.

		This class maintains a set() containing all the event sources
		reverenced in the active branchs of a snapshot manager.

		When a source is added to the collection, it is grabbed by this
		manager (other snapshot manager won't be able tu use it).
		Events occuring within this source will be reported to this
		manager.

		When a source is removed from the collection, the grab is removed.
		"""

		def __init__ (self, manager: "SnapshotManager"):
			self.__set = set()
			self.__manager = manager

		def __iter__ (self):
			return self.__set.__iter__()

		@typecheck
		def update (self, block: SnapshotBlock):
			new_set = set()
			for s in block.get_sources():
				new_set.add (s)

			# grab the new sources
			for s in new_set.difference (self.__set):
				s.set_listener (self.__manager)

			# ungrab the older sources
			for s in self.__set.difference (new_set):
				s.clear_listener (self.__manager)

			self.__set = new_set


		def add (self, source: "EventSource"):
			if source in self.__set:
				return

			self.__set.add (source)
			source.set_listener (self.__manager)

		def clear (self):
			for s in self.__set:
				s.clear_listener (self.__manager)

			self.__set.clear()

	class Branch:
		"""A branch is a path that can be followed in a snapshot block

		A branch is associated to a condition: an Event to be observed
		(possibly with some parameters).

		Optionally the branch can have a handler function, which is
		executed when branch is successfully entered (which implies
		that condition is fulfilled).

		The order in which Branch objects are evaluated is out of the
		scope of this class. It depends on the snapshot block in which
		the are created.

		Branch object are usually never created with an explicit
		instantiation. These objects are created when calling an Event
		object.

		Example:
			with alt:
				# creation of a branch (w/o handler)
				timer.timeout()

				# creation of a branch (w/ handler)
				port.receive(some_message)
				def _():
					print ("message received !)


		"""
		def __init__ (self, event: "Event", *k, **kw):
			self.event = event
			self.k = k
			self.kw = kw
			self.handler = None

		def match (self) -> "optional (EventMatch)":
			return self.event.match (*self.k, **self.kw)

		@typecheck
		def attach_handler (self, func: callable):
			assert self.handler is None
			self.handler = func

	class Thread (threading.Thread):
		"""A thread tied to the local snapshot manager.

		 - its interruption function will notify the local snapshot manager
		 - its status will be tracked by the snapshot manager class
		"""

		__lock = threading.Lock()

		def __init__ (self, *k, **kw):
			threading.Thread.__init__ (self, *k, **kw)

			self.__logger = logger.Logger.get_default()

		def start (self):
			if not self.run is self.__run:
				self.__target = self.run
				self.run = self.__run

			self.__lock.acquire()
			threading.Thread.start (self)

		def __run (self, *k, **kw):
			sm = SnapshotManager()

			self.interrupt = sm.interrupt
			self.__lock.release() # FIXME: releasing in a different thread may not be portable

			with sm.track():
				self.__target(*k, **kw)

		def log_event (self, event: logger.LogEvent):
			self.__logger.log_event (event)

	def __new__ (cls):

		if not hasattr (cls.__thread_local, "manager"):
			self = super (SnapshotManager, cls).__new__(cls)

			self.__thread = threading.currentThread()

			self.__sources = SnapshotManager.Sources (self)
			self.__block = None
			self.__interrupted = False

			self.__lock = threading.Lock()
			self.__cond_update = threading.Condition (self.__lock)

			SnapshotManager.__thread_local.manager = self
		return	SnapshotManager.__thread_local.manager

	def __repr__ (self):
		return "<SnapshotManager %s>" % repr (self.__thread)

	def reset (self):
		self.__sources.clear()
		self.__block = None
		self.__interrupted = False

	def interrupt (self):
		with self.__lock:
			self.__interrupted = True
			self.__cond_update.notify()

			self.__track_wake_up()

	@contextmanager
	def track (self):
		with self.__tracked_lock:
			assert self not in self.__tracked_managers
			self.__tracked_managers[self] = True
		try:
			yield
		finally:
			with self.__tracked_lock:
				assert self in self.__tracked_managers
				del self.__tracked_managers[self]

	def __track_wake_up (self):
		with self.__tracked_lock:
			if self in self.__tracked_managers:
				self.__tracked_managers[self] = True

	def __track_wait (self):
		with self.__tracked_lock:
			if self not in self.__tracked_managers:
				return

			self.__tracked_managers[self] = False

			if any (self.__tracked_managers.values()):
				return

		# All threads are waiting for external events
		# -> notify the clock
		clock.Clock.get_instance().all_threads_blocked()

	@classmethod
	def all_tracked_threads_blocked (cls):
		with cls.__tracked_lock:
			return not any (cls.__tracked_managers.values())

	def run (self) -> "optional(EventMatch)":

		self.__obsolete = True # FIXME: not very clean to touch it when not locked

		try:
			while True:
				with self.__lock:
					if self.__interrupted:
						raise exceptions.UserInterrupt()

					# update the list of sources
					self.__sources.update (self.__block)

					# possibly wait for new events
					if not self.__obsolete:

						self.__track_wait()

#						print "wait"
						self.__cond_update.wait()

						if self.__interrupted:
							raise exceptions.UserInterrupt()

					# make the snapshot
					for s in self.__sources:
						s.evaluate_snapshot()

					self.__obsolete = False

				try:
					event_match = self.__block.match()
					if event_match:
						# force the reevaluation of the snapshot (we may have received 2 events at once)
						self.__obsolete = True # FIXME: not very clean to touch it when not locked

						event_match.call_handler()
						return event_match
				except SnapshotManager.Repeat:
#					print "repeat"
					pass
		finally:
			# clear the source list
			self.__sources.clear()
			pass

	def push_event (self, event: "Event", *k, **kw) -> callable:
		# create a branch to handle that event
		branch = SnapshotManager.Branch (event, *k, **kw)

		if self.__block and self.__block.__editable:
			# we are in an editable block

			# append it to the list of the current block
			self.__block.branch_list.append (branch)

			# return a function to allow attaching a handler to the
			# branch statement
			return branch.attach_handler
		else:
			# the block is not editable
			# -> assume that we have to process the instruction now

			# alt block is the default
			self.push_block (AltBlock())
			try:
				self.__block.branch_list.append (branch)

				event_match = self.run()
				if event_match and event_match.get_branch() == branch:
					def result (func):
						func()
				else:
					def result (func):
						pass
				return result
			finally:
				self.pop_block()

	def lock (self):
		self.__lock.acquire()

	def unlock (self):
		self.__obsolete = True
		self.__cond_update.notify()
		self.__track_wake_up()
		self.__lock.release()

	@typecheck
	def push_block (self, block: SnapshotBlock, editable: bool = False, fail_if_editing: bool = True):

		if fail_if_editing and self.__block:
			assert not self.__block.__editable # TODO: document this error (needed to avoid breaking things with an altstep definition inside another block TODO: check that this assertion is correct


		if self.__block:
			self.__block = self.__block.push (block)
		else:
			self.__block = block

		self.__block.__editable = editable

	@typecheck
	def pop_block (self, keep_editable: bool = False):
		assert self.__block

		self.__block = self.__block.pop()

		if self.__block and not keep_editable:
			self.__block.__editable = False
def repeat():
	raise SnapshotManager.Repeat()

class SnapshotContext:
	def __init__ (self, cls):
		self.__cls = cls

	def __enter__ (self):
		SnapshotManager().push_block (self.__cls(), editable = True)

	def __exit__ (self, exc_type, exc_value, traceback):
		sm = SnapshotManager()
		try:
			if not exc_type:
				sm.run()
		finally:
			sm.pop_block()

alt = SnapshotContext(AltBlock)


class EventSource:
	class Lock:
		def __init__ (self, source):
			self.__source = source
		def __enter__ (self):
			self.__source._lock()
		def __exit__ (self, a, b, c):
			self.__source._unlock()

	def __init__ (self):
		self.__listener	= None
		self.__lock	= threading.Lock()
		self.lock	= EventSource.Lock (self)

	@typecheck
	def set_listener (self, listener: SnapshotManager):
		with self.__lock:
			assert self.__listener is None  # TODO: report an error if not

			self.__listener = listener

	@typecheck
	def clear_listener (self, listener: SnapshotManager):
		with self.__lock:
			assert self.__listener == listener

			self.__listener = None

	def _lock (self):
		self.__lock.acquire()
		if self.__listener:
			self.__listener.lock()

	def _unlock (self):
		if self.__listener:
			self.__listener.unlock()
		self.__lock.release()

class Event:
	@typecheck
	def __init__ (self, source: EventSource, name: str):
		self.__source = source

		# TODO: check that the match function exists
		self.__match_func = getattr (source, "match_" + name)

	@typecheck
	def __call__ (self, *k, **kw) -> either (nothing, callable):
		if kw or len(k) != 1 or type (k[0]) != types.FunctionType:
			# event called with parameters
			return SnapshotManager().push_event (self, *k, **kw)
		else:
			# event called without any parameter
			# (just used to tag the function)
			# FIXME: might not be a good idea (because this function would not be called if it is not a tag)
			return SnapshotManager().push_event (self) (*k, **kw)

	@typecheck
	def get_sources (self) -> iterable:
		yield self.__source

	@typecheck
	def get_source (self) -> EventSource:
		return self.__source

	def match (self, *k, **kw) -> "optional (EventMatch)":
		result = self.__match_func (*k, **kw)
		assert result is None or isinstance (result, EventMatch)
		return result

class EventMatch:
	def __init__ (self, event):
		self.__event = event
		self.__branch = None

	@typecheck
	def set_branch (self, branch: SnapshotManager.Branch):
		self.__branch = branch

	@typecheck
	def get_branch (self) -> optional (SnapshotManager.Branch):
		return self.__branch

	def get_event (self):
		return self.__event

	def get_source (self):
		return self.__event.get_source()

	def call_handler (self):
		#TODO: support any callable object
		#TODO: support *k and **kw

		assert self.__branch is not None  # FIXME: not sure about this one

		func = self.__branch.handler

		if func is None:
			# no handler set
			# -> just return
			return None

		self.__branch = None # reset the branch (to ensure that the handler is not called twice)

		nb = func.__code__.co_argcount
		if not nb:
			return func()
		else:
			params = []
			for name in func.__code__.co_varnames[0:nb]:
				assert hasattr (self, "get_" + name)  # unknown parameter in event callback
				params.append (getattr (self, "get_" + name) ())

			return func (*params)


class Timer (EventSource, logger.LoggedObject, named.NamedObject):

	def __init__ (self, duration = None):
		EventSource.__init__ (self)
		logger.LoggedObject.__init__ (self)
		named.NamedObject.__init__ (self)

		self.timeout = Event (self, "timeout")

		self.__clock = None
		self.__running = False

		if duration is not None:
			self.start (duration)

	def __str__ (self):
		return "timer:%s" % self.__name__

	def start (self, duration):
		with self.lock:
			# cancel the previous timer if any
			if self.__clock:
				self.__clock.unschedule_events (self.__callback)

			# start a new timer
			self.__running = True
			self.__clock   = clock.Clock.get_instance()

			self.__clock.schedule_event_relative (duration, self.__callback)

		self.log (EventTimerStarted, duration)


	def stop (self):
		with self.lock:
			if self.__clock:
				self.__clock.unschedule_events (self.__callback)
				self.__clock = None
				self.__running = False

		self.log (EventTimerStopped)

	def evaluate_snapshot (self):
		self.__snapshot_running = self.__running

	@typecheck
	def match_timeout (self) -> optional (EventMatch):
		if self.__snapshot_running or not self.__clock:
			return None

		self.__clock = None

		self.log (EventTimerTimeout)
		return EventMatch (self.timeout)

	def __callback (self):
		with self.lock:
#			print "timeout callback", self
			self.__running = False


class EventTimerStarted (metaclass = logger.LogEventClass):
	fields = (("timer", Timer),
		  ("duration", numbers.Number),
		 )

	def summary (self):
		return "timer %s started: %s seconds" % (self[0].__name__, self[1])

class EventTimerStopped (metaclass = logger.LogEventClass):
	fields = (("timer", Timer),)

	def summary (self):
		return "timer %s stopped" % self[0].__name__

class EventTimerTimeout (metaclass = logger.LogEventClass):
	fields = (("timer", Timer),)

	def summary (self):
		return "timer %s timeout" % self[0].__name__

# TODO: write a generic class that handles the complex parts of Altstep so that it can be reused for other purposes
class Altstep:
	@typecheck
	def __init__ (self, func: types.FunctionType):
		self.__func = func

	@typecheck
	def __call__ (self, *k, **kw) -> either (nothing, callable):
		block = AltstepBlock (self.__func, *k, **kw)

		# register the event in the snapshot manager
		return block.event()

	#TODO: implement AltstepBlock.deactivate()
	def activate (self, *k, **kw) -> "AltstepBlock":
		block = AltstepBlock (self.__func, *k, **kw)

		SnapshotManager().push_block(block)

		return block

	@staticmethod
	@typecheck
	def create (func: types.FunctionType):
		# create the altstep event source
		return Altstep (func)


class AltstepBlock (EventSource, AltBlock):

	@typecheck
	def __init__ (self, func: types.FunctionType, *k, **kw):
		EventSource.__init__(self)
		AltBlock.__init__(self, func.__name__)

		# Populate the altstep with events
		sm = SnapshotManager()
		sm.push_block (self, editable = True, fail_if_editing = False)
		try:
#			print " running code", func, k, kw
			func (*k, **kw)
		finally:
			sm.pop_block (keep_editable = True)

		self.event = Event (self, "altstep")
		self.event.get_sources = self.get_sources

	@typecheck
	def match_altstep (self) -> optional (EventMatch):
		event_match = self.match()

		if event_match:
			# run it (FIXME: might not be the right place to do this)
			event_match.call_handler()

		return event_match

altstep = Altstep.create

